﻿using C1.TextParser;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace ProductDetailFetcher
{
    public abstract class BaseProductDetailsFetcher
    {
        private HttpClient _client;
        private IXPathPool _xPathPool;
        public IXPathPool XPathPool => _xPathPool;
        protected HttpClient Client => _client;

        public BaseProductDetailsFetcher(IXPathPool xPathPool, IClientConfig config)
        {
            _xPathPool = xPathPool;
            InitClient(config);
        }

        /// <summary>
        /// Intializes the HttpClient responsible for reading the web page source.
        /// </summary>
        /// <param name="config"></param>
        private void InitClient(IClientConfig config)
        {
            HttpClientHandler handler = new HttpClientHandler()
            {
                AutomaticDecompression = config.DecompressionMethod
            };

            _client = new HttpClient(handler)
            {
                BaseAddress = new Uri(config.BaseUri),
            };
        }

        /// <summary>
        /// Gets the product details object of the provided type.<br/>
        /// Note : Valid object type with property names matching the HtmlExtractor placeholders is expected.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="productUri"></param>
        /// <returns></returns>
        public async Task<T> GetProductDetailsAsync<T>(string productUri)
        {
            var html = await Client.ReadPageContentAsync(productUri);
            var extractor = GetExtractor(html);
            var result = extractor.Extract(html.ToStream());     
            GC.Collect();
            return result.ToObject<T>();
        }

        /// <summary>
        /// Gets the extractor using the best matching xpaths of html string.
        /// </summary>
        /// <param name="html"></param>
        /// <returns></returns>
        private IExtractor GetExtractor(string html)
        {
            var xPathObjects = _xPathPool.GetBestMatchingXPaths(html);
            var extractor = CreateExtractor(xPathObjects, html);
            return extractor;
        }
        
        /// <summary>
        /// Creates the extractor for extracting text from source.
        /// </summary>
        /// <param name="template"></param>
        /// <returns></returns>
        protected abstract IExtractor CreateExtractor(IEnumerable<XPathObject> xPathObjects, string html);
    }
}
